__all__ = ('SubscriptionStatus',)

from ...bases import Preinstance as P, PreinstancedBase


class SubscriptionStatus(PreinstancedBase):
    """
    Represents a subscription's status.
    
    Attributes
    ----------
    name : `str`
        The name of the subscription status.
    value : `int`
        The identifier value the subscription status.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``SubscriptionStatus``) items
        Stores the predefined ``SubscriptionStatus``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The subscription status' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the subscription statuses.
    
    Every predefined subscription status can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | Name          | Value |
    +=======================+===============+=======+
    | active                | active        | 0     |
    +-----------------------+---------------+-------+
    | ending                | ending        | 1     |
    +-----------------------+---------------+-------+
    | inactive              | inactive      | 2     |
    +-----------------------+---------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    active = P(0, 'active')
    ending = P(1, 'ending')
    inactive = P(2, 'inactive')



__all__ = ('HangType',)

from ...bases import Preinstance as P, PreinstancedBase


ACTIVITY_NAME_HANGING_DEFAULT = 'Right now, I\'m -'


def _name_getter_custom(activity):
    """
    Name getter for ``HangType`` for its custom instance.
    
    Parameters
    ----------
    activity : ``Activity``
        The activity to get its name of.
    
    Returns
    -------
    name : `str`
    """
    details = activity.details
    emoji = activity.emoji
    if (details is None):
        if (emoji is None):
            name = ACTIVITY_NAME_HANGING_DEFAULT
        else:
            name = emoji.as_emoji
    else:
        if (emoji is None):
            name = details
        else:
            name = f'{emoji} {details}'

    return name


class HangType(PreinstancedBase):
    """
    Represents a hanging activity' type.
    
    Attributes
    ----------
    name : `str`
        The hang type's name.
    
    name_getter : `(Activity) -> str`
        Function used when getting the hanging activity's name.
    
    value : `str`
        The Discord side identifier value of the hang type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``HangType``) items
        Stores the predefined ``HangType``-s.
    
    VALUE_TYPE : `type` = `str`
        The hang types' values' type.
    
    DEFAULT_NAME : `str` = `''`
        The default name of the hang types. Guild features have the same value as name, so at their case it is not
        applicable.
    
    Every predefined hang type can be accessed as class attribute as well:
    
    +-----------------------+---------------+---------------+
    | Class attribute names | name          | value         |
    +=======================+===============+===============+
    | none                  | none          |               |
    +-----------------------+---------------+---------------+
    | be_right_back         | be right back | brb           |
    +-----------------------+---------------+---------------+
    | chilling              | chilling      | chilling      |
    +-----------------------+---------------+---------------+
    | custom                | custom        | custom        |
    +-----------------------+---------------+---------------+
    | focusing              | focusing      | focusing      |
    +-----------------------+---------------+---------------+
    | gaming                | gaming        | gaming        |
    +-----------------------+---------------+---------------+
    | eating                | eating        | eating        |
    +-----------------------+---------------+---------------+
    | travelling            | travelling    | in-transit    |
    +-----------------------+---------------+---------------+
    | watching              | watching      | watching      |
    +-----------------------+---------------+---------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    DEFAULT_NAME_GETTER = lambda activity: ACTIVITY_NAME_HANGING_DEFAULT
    
    __slots__ = ('name_getter', )
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new hang type.
        
        Parameters
        ----------
        value : `str`
            The hang type's identifier value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value.casefold().replace('-', ' ')
        self.name_getter = cls.DEFAULT_NAME_GETTER
        self.INSTANCES[value] = self
        return self
    
    
    def __init__(self, value, name, name_getter):
        """
        Creates a hang type and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `str`
            The Discord side identifier value of the hang type.
        
        name : `str`
            The name of the hang type.
        
        metadata_type : `FunctionType`
            The hang type's name getter.
        """
        self.name = name
        self.name_getter = name_getter
        self.value = value
        
        self.INSTANCES[value] = self
    
    
    # predefined
    none = P('', 'none', DEFAULT_NAME_GETTER)
    be_right_back = P('brb', 'be right back', (lambda activity: 'Gonna BRB'))
    chilling = P('chilling', 'chilling', (lambda activity: 'Chilling'))
    custom = P('custom', 'custom', _name_getter_custom)
    eating = P('eating', 'eating', (lambda activity: 'Grubbin'))
    focusing = P('focusing', 'focusing', (lambda activity: 'In the zone'))
    gaming = P('gaming', 'gaming', (lambda activity: 'GAMING'))
    travelling = P('in-transit', 'travelling', (lambda activity: 'Wandering IRL'))
    watching = P('watching', 'watching', (lambda activity: 'Watchin\' stuff'))



__all__ = ('EmbeddedActivityLocationType', )

from ...bases import Preinstance as P, PreinstancedBase


class EmbeddedActivityLocationType(PreinstancedBase):
    """
    Represents an embedded activity location's type.
    
    Attributes
    ----------
    name : `str`
        The embedded activity location type's name.
    
    value : `str`
        The Discord side identifier value of the embedded activity location type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``EmbeddedActivityLocationType``) items
        Stores the predefined ``EmbeddedActivityLocationType``-s.
    
    VALUE_TYPE : `type` = `str`
        The embedded activity location types' values' type.
    
    DEFAULT_NAME : `str` = `''`
        The default name of the embedded activity location types. Guild features have the same value as name, so at their case it is not
        applicable.
    
    Every predefined embedded activity location type can be accessed as class attribute as well:
    
    +-----------------------+-------------------+---------------+
    | Class attribute names | name              | value         |
    +=======================+===================+===============+
    | none                  | none              |               |
    +-----------------------+-------------------+---------------+
    | guild_channel         | guild channel     | gc            |
    +-----------------------+-------------------+---------------+
    | private_channel       | private channel   | pc            |
    +-----------------------+-------------------+---------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    DEFAULT_NAME = ''
    
    __slots__ = ()
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new embedded activity location type.
        
        Parameters
        ----------
        value : `str`
            The embedded activity location type's identifier value.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.value = value
        self.name = value.casefold().replace('-', ' ')
        self.INSTANCES[value] = self
        return self
    
    
    # predefined
    none = P('', 'none')
    guild_channel = P('gc', 'guild channel')
    private_channel = P('pc', 'private channel')



__all__ = (
    'API_ENDPOINT', 'CDN_ENDPOINT', 'DISCORD_ENDPOINT', 'INVITE_URL_RP', 'STATUS_ENDPOINT', 'VALID_ICON_FORMATS',
    'VALID_ICON_FORMATS_EXTENDED', 'VALID_ICON_MEDIA_TYPES', 'VALID_ICON_MEDIA_TYPES_EXTENDED',
    'VALID_IMAGE_MEDIA_TYPES_ALL', 'VALID_STICKER_IMAGE_MEDIA_TYPES', 'is_media_url', 'parse_message_jump_url'
)

import re

from scarletio import export, include

from ...env import (
    API_VERSION, CUSTOM_API_ENDPOINT, CUSTOM_CDN_ENDPOINT, CUSTOM_DISCORD_ENDPOINT, CUSTOM_MEDIA_ENDPOINT,
    CUSTOM_STATUS_ENDPOINT
)


StickerFormat = include('StickerFormat')
_try_get_guild_id = include('_try_get_guild_id')


API_ENDPOINT = f'https://discord.com/api/v{API_VERSION}' if (CUSTOM_API_ENDPOINT is None) else CUSTOM_API_ENDPOINT
CDN_ENDPOINT = 'https://cdn.discordapp.com' if (CUSTOM_CDN_ENDPOINT is None) else CUSTOM_CDN_ENDPOINT
DISCORD_ENDPOINT = 'https://discord.com' if (CUSTOM_DISCORD_ENDPOINT is None) else CUSTOM_DISCORD_ENDPOINT
STATUS_ENDPOINT = 'https://status.discord.com/api/v2' if (CUSTOM_STATUS_ENDPOINT is None) else CUSTOM_STATUS_ENDPOINT
MEDIA_ENDPOINT =  'https://media.discordapp.net' if (CUSTOM_MEDIA_ENDPOINT is None) else CUSTOM_MEDIA_ENDPOINT

del CUSTOM_API_ENDPOINT, CUSTOM_CDN_ENDPOINT, CUSTOM_DISCORD_ENDPOINT, CUSTOM_STATUS_ENDPOINT, API_VERSION

VALID_ICON_SIZES = frozenset((
    *( 1 << x      for x in range(4, 13)),
    *((1 << x) * 3 for x in range(9, 11)),
    *((1 << x) * 5 for x in range(2,  9)),
))

VALID_ICON_FORMATS = frozenset(('jpg', 'jpeg', 'png', 'webp'))
VALID_ICON_FORMATS_EXTENDED = frozenset((*VALID_ICON_FORMATS, 'gif',))

VALID_ICON_MEDIA_TYPES = frozenset(('image/jpeg', 'image/png', 'image/webp'))
VALID_ICON_MEDIA_TYPES_EXTENDED = frozenset(('image/gif', *VALID_ICON_MEDIA_TYPES))

VALID_STICKER_IMAGE_MEDIA_TYPES = frozenset(('image/gif', 'image/png', 'application/json'))
VALID_IMAGE_MEDIA_TYPES_ALL = frozenset((*VALID_ICON_MEDIA_TYPES_EXTENDED, *VALID_STICKER_IMAGE_MEDIA_TYPES))


WIDGET_STYLE_RP = re.compile('shield|banner[1-4]')

MESSAGE_JUMP_URL_RP = re.compile(
    '(?:https://)?(?:(?:canary|ptb)\\.)?discord(?:app)?.com/channels/(?:(\\d{7,21})|@me)/(\\d{7,21})/(\\d{7,21})'
)
export(MESSAGE_JUMP_URL_RP, 'MESSAGE_JUMP_URL_RP')


def _validate_extension(icon_type, ext):
    """
    Validates the given icon extension.
    
    Parameters
    ----------
    icon_type : ``IconType``
        The respective icon type.
    ext : `None`, `str`
        The received extension.
    
    Returns
    -------
    ext : `str`
        The validated extension.
    
    Raises
    ------
    ValueError
        - If `ext`'s value is not applicable for the given icon type.
    """
    if ext is None:
        ext = icon_type.default_postfix
    
    else:
        if not icon_type.allows_postfix(ext):
            raise ValueError(
                f'Extension must be one of {ext.allowed_postfixes}, got {ext!r}.'
            )
    
    return ext


def _build_end(size):
    """
    Validates the given icon size.
    
    Parameters
    ----------
    size : `None`, `int`
        The received size.
    
    Returns
    -------
    end : `str`
        The validated size as query string.
    
    Raises
    ------
    ValueError
        - If `size`'s value is not applicable for the given icon type.
    """
    if size is None:
        end = ''
    
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    
    else:
        raise ValueError(
            f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size!r}.'
        )
    
    return end


# returns a URL that allows the client to jump to this message
# guild is guild's id, or @me if there is no guild
def message_jump_url(message):
    """
    Returns a jump url to the message. If the message's channel is a partial guild channel, returns `None`.
    
    This function is a shared property of ``Message``-s.
    
    Returns
    -------
    url : `str`
    """
    channel_id = message.channel_id
    guild_id = message.guild_id
    if guild_id:
        guild_id = str(guild_id)
    else:
        guild_id = '@me'
    
    return f'{DISCORD_ENDPOINT}/channels/{guild_id}/{channel_id}/{message.id}'


def parse_message_jump_url(message_url):
    """
    Parses the jump url of a message. On failure returns `0`-s.
    
    Parameters
    ----------
    message_url : `str`
        The message url to parse.
    
    Returns
    -------
    guild_id : `int`
        The message's guild's identifier. Defaults to `0` if the message is from a private channel.
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    """
    parsed = MESSAGE_JUMP_URL_RP.fullmatch(message_url)
    if parsed is None:
        guild_id = 0
        channel_id = 0
        message_id = 0
    else:
        guild_id, channel_id, message_id = parsed.groups()
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        channel_id = int(channel_id)
        message_id = int(message_id)
    
    return guild_id, channel_id, message_id


CDN_RP = re.compile(
    'https://(?:'
        'cdn\\.discordapp\\.com|'
        '(?:(?:canary|ptb)\\.)?discord\\.com|'
        '(?:'
            'images-ext-\\d+|'
            'media'
        ')\\.discordapp\\.net'
    ')/'
)

def is_cdn_url(url):
    """
    Returns whether the given url a Discord content delivery network url.
    
    Parameters
    ----------
    url : `str`
        The url to check.
    
    Returns
    -------
    is_cdn_url : `bool`
    
    Examples
    --------
    Icons: `https://cdn.discordapp.com/...`
    Assets: `https://discord.com/...`
    Proxy service: `https://images-ext-1.discordapp.net/...`
    Attachments: `https://media.discordapp.net/...`
    ```
    """
    return (CDN_RP.match(url) is not None)


def is_media_url(url):
    """
    Returns whether the given url uses the discord's media content delivery network.
    
    Parameters
    ----------
    url : `str`
        The url to check.
    
    Returns
    -------
    is_media_url : `bool`
    """
    return url.startswith('https://media.discordapp.net/')


def guild_icon_url(guild):
    """
    Returns the guild's icon's image's url. If the guild has no icon, then returns `None`.
    
    This function is a shared property of ``Guild``, ``GuildPreview``.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = guild.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/icons/{guild.id}/{prefix}{guild.icon_hash:0>32x}.{ext}'


def guild_icon_url_as(guild, ext = None, size = None):
    """
    Returns the guild's icon's url. If the guild has no icon, then returns `None`.
    
    This function is a shared method of ``Guild``, ``GuildPreview``.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the guild has
        animated icon, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = guild.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/icons/{guild.id}/{prefix}{guild.icon_hash:0>32x}.{ext}{end}'


def guild_invite_splash_url(guild):
    """
    Returns the guild's invite splash's image's url. If the guild has no invite splash, then returns `None`.
    
    This function is a shared property of ``Guild``, ``GuildPreview``.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = guild.invite_splash_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/splashes/{guild.id}/{prefix}{guild.invite_splash_hash:0>32x}.{ext}'


def guild_invite_splash_url_as(guild, ext = None, size = None):
    """
    Returns the guild's invite splash's image's url. If the guild has no invite splash, then returns `None`.
    
    This function is a shared method of ``Guild``, ``GuildPreview``.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = guild.invite_splash_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/splashes/{guild.id}/{prefix}{guild.invite_splash_hash:0>32x}.{ext}{end}'


def guild_discovery_splash_url(guild):
    """
    Returns the guild's discovery splash's image's url. If the guild has no discovery splash, then returns `None`.
    
    This function is a shared property of ``Guild``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = guild.discovery_splash_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/discovery-splashes/{guild.id}/{prefix}{guild.discovery_splash_hash:0>32x}.{ext}'


def guild_discovery_splash_url_as(guild, ext = None, size = None):
    """
    Returns the guild's discovery splash's image's url. If the guild has no discovery splash, then returns `None`.
    
    This function is a shared method of ``Guild``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = guild.discovery_splash_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/discovery-splashes/{guild.id}/{prefix}{guild.discovery_splash_hash:0>32x}.{ext}{end}'


def guild_banner_url(guild):
    """
    Returns the guild's banner's image's url. If the guild has no banner, then returns `None`.
    
    This function is a shared property of ``Guild``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = guild.banner_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/banners/{guild.id}/{prefix}{guild.banner_hash:0>32x}.{ext}'


def guild_banner_url_as(guild, ext = None, size = None):
    """
    Returns the guild's banner's image's url. If the guild has no banner, then returns `None`.
    
    This function is a shared method of ``Guild``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`, `'gif'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = guild.banner_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/banners/{guild.id}/{prefix}{guild.banner_hash:0>32x}.{ext}{end}'


def guild_widget_url(guild):
    """
    Returns the guild's widget image's url in `.png` format.
    
    This function is a shared property of ``Guild``-s.
    
    Returns
    -------
    url : `str`
    """
    return f'{API_ENDPOINT}/guilds/{guild.id}/widget.png?style=shield'


def guild_widget_url_as(guild, style = 'shield'):
    """
    Returns the guild's widget image's url in `.png` format.
    
    This function is a shared method of ``Guild``-s.
    
    Parameters
    ----------
    style : `str` = `'shield'`, Optional
        The widget image's style. Can be any of: `'shield'`, `'banner1'`, `'banner2'`, `'banner3'`, `'banner4'`.
    
    Returns
    -------
    url : `str`
    
    Raises
    ------
    ValueError
        If `style` was not passed as any of the expected values.
    """
    if WIDGET_STYLE_RP.fullmatch(style) is None:
        raise ValueError(f'Invalid style: {style!r}')
    
    return f'{API_ENDPOINT}/guilds/{guild.id}/widget.png?style={style}'


def guild_widget_json_url(guild):
    """
    Returns an url to request a ``Guild``'s widget data.
    
    This function is a shared property of ``Guild``, ``GuildWidget``.
    
    Returns
    -------
    url : `str`
    """
    return  f'{API_ENDPOINT}/guilds/{guild.id}/widget.json'


def channel_group_icon_url(channel):
    """
    Returns the group channel's icon's image's url. If the channel has no icon, then returns `None`.
    
    This function is a shared property of ``Channel``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = channel.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/channel-icons/{channel.id}/{prefix}{channel.icon_hash:0>32x}.{ext}'
    
    
def channel_group_icon_url_as(channel, ext = None, size = None):
    """
    Returns the group channel's icon's image's url. If the channel has no icon, then returns `None`.
    
    This function is a shared method of ``Channel``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = channel.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/channel-icons/{channel.id}/{prefix}{channel.icon_hash:0>32x}.{ext}{end}'


def emoji_url(emoji):
    """
    Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
    
    This function is a shared property of ``Emoji``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    if not emoji.is_custom_emoji():
        return None
    
    if emoji.animated:
         ext = 'gif'
    else:
         ext = 'png'
    
    return f'{CDN_ENDPOINT}/emojis/{emoji.id}.{ext}'


def emoji_url_as(emoji, ext = None, size = None):
    """
    Returns the emoji's image's url. If the emoji is unicode emoji, then returns `None` instead.
    
    This function is a shared method of ``Emoji``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If emoji is
        animated, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    if not emoji.is_custom_emoji():
        return None
    
    if ext is None:
        if emoji.animated:
            ext = 'gif'
        else:
            ext = 'png'
    else:
        if emoji.animated:
            if ext not in VALID_ICON_FORMATS_EXTENDED:
                raise ValueError(
                    f'Extension must be one of {VALID_ICON_FORMATS_EXTENDED}, got {ext!r}.'
                )
        else:
            if ext not in VALID_ICON_FORMATS:
                raise ValueError(
                    f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.'
                )
    
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/emojis/{emoji.id}.{ext}{end}'


def webhook_url(webhook):
    """
    Returns the webhook's url.
    
    This function is a shared property of ``Webhook``-s.
    
    Returns
    -------
    url : `str`
    """
    return f'{API_ENDPOINT}/webhooks/{webhook.id}/{webhook.token}'


WEBHOOK_URL_PATTERN = re.compile(
    '(?:https://)?discord(?:app)?.com/api/(?:v\\d+/)?webhooks/([0-9]{17,21})/([a-zA-Z0-9.\\-_%]{60,68})(?:/.*)?'
)


def _build_invite_url(invite_code):
    """
    Builds an invite's url from the given core.
    
    Parameters
    ----------
    invite_code : `str`
        An invite's code.
    
    Returns
    -------
    url : `str`
    """
    return f'http://discord.gg/{invite_code}'


def invite_url(invite):
    """
    Returns the invite's url.
    
    This function is a shared property of ``Invite``-s.
    
    Returns
    -------
    url : `str`
    """
    return _build_invite_url(invite.code)


def guild_vanity_invite_url(guild):
    """
    Returns the guild vanity invite's url.
    
    This function is a shared property of ``Guild``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    vanity_code = guild.vanity_code
    if (vanity_code is not None):
        return _build_invite_url(vanity_code)


INVITE_URL_RP = re.compile('(?:https?://)?discord(?:\\.gg|(?:app)?\\.com/invite)/([a-zA-Z0-9-]+)')


def activity_asset_image_large_url(activity):
    """
    Returns the activity's large asset image's url. If the activity has no large asset image, then returns `None`.
    
    This function is a shared property of ``Activity``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    application_id = activity.application_id
    if not application_id:
        return None
    
    assets = activity.assets
    if assets is None:
        return None
    
    image_large = assets.image_large
    if image_large is None:
        return None
    
    if not image_large.isdigit():
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_large}.png'


def activity_asset_image_large_url_as(activity, ext = None, size = None):
    """
    Returns the activity's large asset image's url. If the activity has no large asset image, then returns `None`.
    
    This function is a shared method of ``Activity``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    application_id = activity.application_id
    if not application_id:
        return None

    assets = activity.assets
    if assets is None:
        return None
    
    image_large = assets.image_large
    if image_large is None:
        return None
    
    if not image_large.isdigit():
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size!r}.')

    if ext is None:
        ext = 'png'
    elif ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_large}.{ext}{end}'


def activity_asset_image_small_url(activity):
    """
    Returns the activity's small asset image's url. If the activity has no small asset image, then returns `None`.
    
    This function is a shared property of ``Activity``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    application_id = activity.application_id
    if not application_id:
        return None
    
    assets = activity.assets
    if assets is None:
        return None
    
    image_small = assets.image_small
    if image_small is None:
        return None
    
    if not image_small.isdigit():
        return None
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_small}.png'


def activity_asset_image_small_url_as(activity, ext = None, size = None):
    """
    Returns the activity's small asset image's url. If the activity has no small asset image, then returns `None`.
    
    This function is a shared method of ``Activity``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    application_id = activity.application_id
    if not application_id:
        return None
    
    assets = activity.assets
    if assets is None:
        return None
    
    image_small = assets.image_small
    if image_small is None:
        return None
    
    if not image_small.isdigit():
        return None
    
    if size is None:
        end = ''
    elif size in VALID_ICON_SIZES:
        end = f'?size={size}'
    else:
        raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size!r}.')
    
    if ext is None:
        ext = 'png'
    elif ext not in VALID_ICON_FORMATS:
        raise ValueError(f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.')
    
    return f'{CDN_ENDPOINT}/app-assets/{application_id}/{image_small}.{ext}{end}'


def user_banner_url(user):
    """
    Returns the user's banner's url. If the user has no banner, then returns `None`.
    
    This function is a shared property of ``UserBase``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = user.banner_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/banners/{user.id}/{prefix}{user.banner_hash:0>32x}.{ext}'


def user_banner_url_as(user, ext = None, size = None):
    """
    Returns the user's banner's url. If the user has no banner, then returns `None`.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated banner, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = user.banner_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/banners/{user.id}/{prefix}{user.banner_hash:0>32x}.{ext}{end}'


def user_banner_url_for(user, guild):
    """
    Returns the user's guild specific banner. If the user has no guild local banner, returns `None`.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    
    Returns
    -------
    url : `None`, `str`
    """
    guild_id = _try_get_guild_id(guild)
    
    try:
        guild_profile = user.guild_profiles[guild_id]
    except KeyError:
        return None
    
    icon_type = guild_profile.banner_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user.id}/banners/{prefix}{guild_profile.banner_hash:0>32x}.{ext}'


def user_banner_url_for_as(user, guild, ext = None, size = None):
    """
    Returns the user's guild specific banner. If the user has no guild local banner, then returns `None`.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'` and `'gif'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the banner's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    guild_id = _try_get_guild_id(guild)
    
    try:
        guild_profile = user.guild_profiles[guild_id]
    except KeyError:
        return None
    
    icon_type = guild_profile.banner_type
    if not icon_type.can_create_url():
        return None

    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return (
        f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user.id}/banners/{prefix}{guild_profile.banner_hash:0>32x}.{ext}{end}'
    )


def user_banner_url_at(user, guild):
    """
    Returns the user's banner's url at the guild.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    
    Returns
    -------
    url : `None`, `str`
    """
    banner_url = user_banner_url_for(user, guild)
    if banner_url is None:
        banner_url = user_banner_url(user)
    
    return banner_url


def user_banner_url_at_as(user, guild, ext = None, size = None):
    """
    Returns the user's banner's url at the guild. If the user has no banner, then returns it's default banner's url.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'` and `'gif'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the banner's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    banner_url = user_banner_url_for_as(user, guild, ext = ext, size = size)
    if banner_url is None:
        banner_url = user_banner_url_as(user, ext = ext, size = size)
    
    return banner_url


def user_avatar_url(user):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns it's default avatar's url.
    
    This function is a shared property of ``UserBase``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = user.avatar_type
    if not icon_type.can_create_url():
        return user.default_avatar.url
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/avatars/{user.id}/{prefix}{user.avatar_hash:0>32x}.{ext}'


def user_avatar_url_as(user, ext = None, size = None):
    """
    Returns the user's avatar's url. If the user has no avatar, then returns it's default avatar's url.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated avatar, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = user.avatar_type
    if not icon_type.can_create_url():
        return user.default_avatar.url
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/avatars/{user.id}/{prefix}{user.avatar_hash:0>32x}.{ext}{end}'


def user_avatar_url_for(user, guild):
    """
    Returns the user's guild specific avatar. If the user has no guild local avatar, returns `None`.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    
    Returns
    -------
    url : `None`, `str`
    """
    guild_id = _try_get_guild_id(guild)
    
    try:
        guild_profile = user.guild_profiles[guild_id]
    except KeyError:
        return None
    
    icon_type = guild_profile.avatar_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user.id}/avatars/{prefix}{guild_profile.avatar_hash:0>32x}.{ext}'


def user_avatar_url_for_as(user, guild, ext = None, size = None):
    """
    Returns the user's guild specific avatar. If the user has no guild local avatar, then returns `None`.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated avatar, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    guild_id = _try_get_guild_id(guild)
    
    try:
        guild_profile = user.guild_profiles[guild_id]
    except KeyError:
        return None
    
    icon_type = guild_profile.avatar_type
    if not icon_type.can_create_url():
        return None

    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return (
        f'{CDN_ENDPOINT}/guilds/{guild_id}/users/{user.id}/avatars/{prefix}{guild_profile.avatar_hash:0>32x}.{ext}{end}'
    )


def user_avatar_url_at(user, guild):
    """
    Returns the user's avatar's url at the guild.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    
    Returns
    -------
    url : `None`, `str`
    """
    avatar_url = user_avatar_url_for(user, guild)
    if avatar_url is None:
        avatar_url = user_avatar_url(user)
    
    return avatar_url


def user_avatar_url_at_as(user, guild, ext = None, size = None):
    """
    Returns the user's avatar's url at the guild. If the user has no avatar, then returns it's default avatar's url.
    
    This function is a shared method of ``UserBase``-s.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The respective guild or it's identifier.
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the user has
        animated avatar, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the avatar's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    avatar_url = user_avatar_url_for_as(user, guild, ext = ext, size = size)
    if avatar_url is None:
        avatar_url = user_avatar_url_as(user, ext = ext, size = size)
    
    return avatar_url


def default_avatar_url(default_avatar):
    """
    Returns the user's default avatar's url.
    
    This function is a shared property of ``UserBase``-s.
    
    Returns
    -------
    url : `str`
    """
    return f'{CDN_ENDPOINT}/embed/avatars/{default_avatar.value}.png'


def application_icon_url(application):
    """
    Returns the application's icon's url. If the application has no icon, then returns `None`.
    
    This function is a shared property of ``Application``, ``MessageApplication``, ``IntegrationApplication``.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = application.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/app-icons/{application.id}/{prefix}{application.icon_hash:0>32x}.{ext}'


def application_icon_url_as(application, ext = None, size = None):
    """
    Returns the application's icon's url. If the application has no icon, then returns `None`.
    
    This function is a shared method of ``Application``, ``MessageApplication``, ``IntegrationApplication``.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = application.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/app-icons/{application.id}/{prefix}{application.icon_hash:0>32x}.{ext}{end}'


def application_cover_url(application):
    """
    Returns the application's cover image's url. If the application has no cover image, then returns `None`.
    
    This function is a shared property of ``Application``, ``MessageApplication``.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = application.cover_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/app-assets/{application.id}/store/{prefix}{application.cover_hash:0>32x}.{ext}'


def application_cover_url_as(application, ext = None, size = None):
    """
    Returns the application's cover image's url. If the application has no cover image, then returns `None`.
    
    This function is a shared method of ``Application``, ``MessageApplication``.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the cover's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the cover's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = application.cover_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/app-assets/{application.id}/store/{prefix}{application.cover_hash:0>32x}.{ext}{end}'


def team_icon_url(team):
    """
    Returns the team's icon's url. If the team has no icon, then returns `None`.
    
    This function is a shared property of ``Team``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = team.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/team-icons/{team.id}/{prefix}{team.icon_hash:0>32x}.{ext}'


def team_icon_url_as(team, ext = None, size = None):
    """
    Returns the team's icon's url. If the team has no icon, then returns `None`.
    
    This function is a shared method of ``Team``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = team.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/team-icons/{team.id}/{prefix}{team.icon_hash:0>32x}.{ext}{end}'


def achievement_icon_url(achievement):
    """
    Returns the achievement's icon's url.
    
    This function is a shared property of ``Achievement``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = achievement.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/app-assets/{achievement.application_id}/achievements/{achievement.id}/icons/{prefix}' \
           f'{achievement.icon_hash:0>32x}.{ext}'


def achievement_icon_url_as(achievement, ext = None, size = None):
    """
    Returns the achievement's icon's url.
    
    This function is a shared method of ``Achievement``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the icon's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the icon's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = achievement.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return (
        f'{CDN_ENDPOINT}/app-assets/{achievement.application_id}/achievements/{achievement.id}/icons/{prefix}'
       f'{achievement.icon_hash:0>32x}.{ext}{end}'
    )


def sticker_url(sticker):
    """
    Returns the sticker's url.
    
    This function is a shared method of ``property``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    sticker_format = sticker.format
    if sticker_format is StickerFormat.none:
        return None
    
    if sticker_format is StickerFormat.gif:
        endpoint = MEDIA_ENDPOINT
    else:
        endpoint = CDN_ENDPOINT
        
    return f'{endpoint}/stickers/{sticker.id}.{sticker_format.extension}'


def sticker_url_as(sticker, size = None, preview = False):
    """
    Returns the sticker's url.
    
    This function is a shared method of ``Sticker``-s.
    
    Parameters
    ----------
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the icon's url.
    preview : `bool` = `False`, Optional
        Whether preview url should be generated.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `size` was not passed as any of the expected values.
    """
    sticker_format = sticker.format
    if sticker_format is StickerFormat.none:
        return None
    
    # Resolve size
    if size is None:
        end = ''
    else:
        if sticker_format is StickerFormat.lottie:
            end = ''
        else:
            if size in VALID_ICON_SIZES:
                end = f'?size={size}'
            else:
                raise ValueError(f'Size must be in {sorted(VALID_ICON_SIZES)!r}, got {size!r}.')
    
    # Resolve preview
    if preview:
        if sticker_format is StickerFormat.apng:
            end = f'{end}{"&" if end else "?"}passthrough=false'
    
    if sticker_format is StickerFormat.gif:
        endpoint = MEDIA_ENDPOINT
    else:
        endpoint = CDN_ENDPOINT
    
    return f'{endpoint}/stickers/{sticker.id}.{sticker_format.extension}{end}'


def sticker_pack_banner(sticker_pack):
    """
    Returns the sticker pack banner's url.
    
    This function is a shared property of ``StickerPack``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    banner_id = sticker_pack.banner_id
    if banner_id:
        return f'{CDN_ENDPOINT}/app-assets/710982414301790216/store/{banner_id}.png'


def sticker_pack_banner_as(sticker_pack, ext = None, size = None):
    """
    Returns the achievement's icon's url.
    
    This function is a shared method of ``StickerPack``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the banner's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the banner's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    end = _build_end(size)
    
    if ext is None:
        ext = 'png'
    
    else:
        if ext not in VALID_ICON_FORMATS:
            raise ValueError(
                f'Extension must be one of {VALID_ICON_FORMATS}, got {ext!r}.'
            )
    
    return f'{CDN_ENDPOINT}/app-assets/710982414301790216/store/{sticker_pack.banner_id}.{ext}{end}'


def role_icon_url(role):
    """
    Returns the role's icon's image's url. If the role has no icon, then returns `None`.
    
    This function is a shared property of ``Role``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = role.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/role-icons/{role.id}/{prefix}{role.icon_hash:0>32x}.{ext}'


def role_icon_url_as(role, ext = None, size = None):
    """
    Returns the role's icon's image's url. If the role has no icon, then returns `None`.
    
    This function is a shared method of ``Role``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`, `'gif'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = role.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/role-icons/{role.id}/{prefix}{role.icon_hash:0>32x}.{ext}{end}'


def scheduled_event_image_url(scheduled_event):
    """
    Returns the scheduled event's image's url. If the scheduled event has no image, then returns `None`.
    
    This function is a property of ``ScheduledEvent``-s.
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = scheduled_event.image_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/guild-events/{scheduled_event.id}/{prefix}{scheduled_event.image_hash:0>32x}.{ext}'


def scheduled_event_image_url_as(scheduled_event, ext = None, size = None):
    """
    Returns the scheduled event's image's url. If the scheduled event has no image, then returns `None`.
    
    This function is a method of ``ScheduledEvent``-s.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = scheduled_event.image_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/guild-events/{scheduled_event.id}/{prefix}{scheduled_event.image_hash:0>32x}.{ext}{end}'


def scheduled_event_url(scheduled_event):
    """
    Returns the scheduled event's url.
    
    This function is a property of ``ScheduledEvent``-s.
    
    Returns
    -------
    url : `str`
    """
    return f'{DISCORD_ENDPOINT}/events/{scheduled_event.guild_id}/{scheduled_event.id}'


def user_avatar_decoration_url(user_or_guild_profile):
    """
    Returns the user's or guild profile's avatar decoration's url. If the user has no avatar decoration returns `None`.
    
    This function is a property of ``UserBase`` and ``GuildProfile``.
    
    Returns
    -------
    url : `None`, `str`
    """
    avatar_decoration = user_or_guild_profile.avatar_decoration
    if avatar_decoration is None:
        return None

    icon_type = avatar_decoration.asset_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/avatar-decoration-presets/{prefix}{avatar_decoration.asset_hash:0>32x}.{ext}'


def user_avatar_decoration_url_as(user_or_guild_profile, ext = None, size = None):
    """
    Returns the user's or guild profile's avatar decoration's url. If the user has no avatar decoration returns `None`.
    
    This function is a method of ``UserBase`` and ``GuildProfile``.
    
    Returns
    -------
    url : `None`, `str`
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `png`.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    avatar_decoration = user_or_guild_profile.avatar_decoration
    if avatar_decoration is None:
        return None
    
    icon_type = avatar_decoration.asset_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/avatar-decoration-presets/{prefix}{avatar_decoration.asset_hash:0>32x}.{ext}{end}'


def soundboard_sound_url(sound):
    """
    Returns the url to the sound board sound.
    
    This function is a property of ``SoundboardSound``.
    
    Returns
    -------
    url : `str`
    """
    return f'{CDN_ENDPOINT}/soundboard-sounds/{sound.id}'


def clan_icon_url(clan):
    """
    Returns the clan's icon's image's url. If the clan has no icon, then returns `None`.
    
    This function is a property of ``UserClan``
    
    Returns
    -------
    url : `None`, `str`
    """
    icon_type = clan.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = icon_type.default_postfix
    
    return f'{CDN_ENDPOINT}/clan-badges/{clan.guild_id}/{prefix}{clan.icon_hash:0>32x}.{ext}'


def clan_icon_url_as(clan, ext = None, size = None):
    """
    Returns the clan's icon's url. If the clan has no icon, then returns `None`.
    
    This function is a method of ``UserClan``.
    
    Parameters
    ----------
    ext : `None`, `str` = `None`, Optional
        The extension of the image's url. Can be any of: `'jpg'`, `'jpeg'`, `'png'`, `'webp'`. If the clan has
        animated icon, it can `'gif'` as well.
    size : `None`, `int` = `None`, Optional
        The preferred minimal size of the image's url.
    
    Returns
    -------
    url : `None`, `str`
    
    Raises
    ------
    ValueError
        If `ext`, `size` was not passed as any of the expected values.
    """
    icon_type = clan.icon_type
    if not icon_type.can_create_url():
        return None
    
    prefix = icon_type.prefix
    ext = _validate_extension(icon_type, ext)
    end = _build_end(size)
    
    return f'{CDN_ENDPOINT}/clan-badges/{clan.guild_id}/{prefix}{clan.icon_hash:0>32x}.{ext}{end}'



__all__ = ('Permission', )

from datetime import datetime as DateTime, timezone as TimeZone

from warnings import warn

from ..bases import FlagBase, FlagDescriptor as F


DO_CAN_DEPRECATION = DateTime.now(tz = TimeZone.utc) > DateTime(2025, 5, 14, tzinfo = TimeZone.utc)


PERMISSION_SHIFT_CREATE_INSTANT_INVITE = 0
PERMISSION_SHIFT_KICK_USERS = 1
PERMISSION_SHIFT_BAN_USERS = 2
PERMISSION_SHIFT_ADMINISTRATOR = 3
PERMISSION_SHIFT_MANAGE_CHANNELS = 4
PERMISSION_SHIFT_MANAGE_GUILD = 5
PERMISSION_SHIFT_ADD_REACTION = 6
PERMISSION_SHIFT_VIEW_AUDIT_LOGS = 7
PERMISSION_SHIFT_PRIORITY_SPEAKER = 8
PERMISSION_SHIFT_STREAM = 9
PERMISSION_SHIFT_VIEW_CHANNEL = 10
PERMISSION_SHIFT_SEND_MESSAGES = 11
PERMISSION_SHIFT_SEND_TTS_MESSAGES = 12
PERMISSION_SHIFT_MANAGE_MESSAGES = 13
PERMISSION_SHIFT_EMBED_LINKS = 14
PERMISSION_SHIFT_ATTACH_FILES = 15
PERMISSION_SHIFT_READ_MESSAGE_HISTORY = 16
PERMISSION_SHIFT_MENTION_EVERYONE = 17
PERMISSION_SHIFT_USE_EXTERNAL_EMOJIS = 18
PERMISSION_SHIFT_VIEW_GUILD_INSIGHTS = 19
PERMISSION_SHIFT_CONNECT = 20
PERMISSION_SHIFT_SPEAK = 21
PERMISSION_SHIFT_MUTE_USERS = 22
PERMISSION_SHIFT_DEAFEN_USERS = 23
PERMISSION_SHIFT_MOVE_USERS = 24
PERMISSION_SHIFT_USE_VOICE_ACTIVATION = 25
PERMISSION_SHIFT_CHANGE_NICKNAME = 26
PERMISSION_SHIFT_MANAGE_NICKNAMES = 27
PERMISSION_SHIFT_MANAGE_ROLES = 28
PERMISSION_SHIFT_MANAGE_WEBHOOKS = 29
PERMISSION_SHIFT_MANAGE_GUILD_EXPRESSIONS = 30
PERMISSION_SHIFT_USE_APPLICATION_COMMANDS = 31
PERMISSION_SHIFT_REQUEST_TO_SPEAK = 32
PERMISSION_SHIFT_MANAGE_EVENTS = 33
PERMISSION_SHIFT_MANAGE_THREADS = 34
PERMISSION_SHIFT_CREATE_PUBLIC_THREADS = 35
PERMISSION_SHIFT_CREATE_PRIVATE_THREADS = 36
PERMISSION_SHIFT_USE_EXTERNAL_STICKERS = 37
PERMISSION_SHIFT_SEND_MESSAGES_IN_THREADS = 38
PERMISSION_SHIFT_USE_EMBEDDED_ACTIVITIES = 39
PERMISSION_SHIFT_MODERATE_USERS = 40
PERMISSION_SHIFT_VIEW_CREATOR_MONETIZATION_ANALYTICS = 41
PERMISSION_SHIFT_USE_SOUNDBOARD = 42
PERMISSION_SHIFT_CREATE_GUILD_EXPRESSIONS = 43
PERMISSION_SHIFT_CREATE_GUILD_EVENTS = 44
PERMISSION_SHIFT_USE_EXTERNAL_SOUNDS = 45
PERMISSION_SHIFT_SEND_VOICE_MESSAGES = 46
PERMISSION_SHIFT_USE_CLYDE_AI = 47
PERMISSION_SHIFT_SET_VOICE_CHANNEL_STATUS = 48
PERMISSION_SHIFT_SEND_POLLS = 49
PERMISSION_SHIFT_USE_EXTERNAL_APPLICATION_COMMANDS = 50


PERMISSION_MASK_CREATE_INSTANT_INVITE = 1 << PERMISSION_SHIFT_CREATE_INSTANT_INVITE
PERMISSION_MASK_KICK_USERS = 1 << PERMISSION_SHIFT_KICK_USERS
PERMISSION_MASK_BAN_USERS = 1 << PERMISSION_SHIFT_BAN_USERS
PERMISSION_MASK_ADMINISTRATOR = 1 << PERMISSION_SHIFT_ADMINISTRATOR
PERMISSION_MASK_MANAGE_CHANNELS = 1 << PERMISSION_SHIFT_MANAGE_CHANNELS
PERMISSION_MASK_MANAGE_GUILD = 1 << PERMISSION_SHIFT_MANAGE_GUILD
PERMISSION_MASK_ADD_REACTION = 1 << PERMISSION_SHIFT_ADD_REACTION
PERMISSION_MASK_VIEW_AUDIT_LOGS = 1 << PERMISSION_SHIFT_VIEW_AUDIT_LOGS
PERMISSION_MASK_PRIORITY_SPEAKER = 1 << PERMISSION_SHIFT_PRIORITY_SPEAKER
PERMISSION_MASK_STREAM = 1 << PERMISSION_SHIFT_STREAM
PERMISSION_MASK_VIEW_CHANNEL = 1 << PERMISSION_SHIFT_VIEW_CHANNEL
PERMISSION_MASK_SEND_MESSAGES = 1 << PERMISSION_SHIFT_SEND_MESSAGES
PERMISSION_MASK_SEND_TTS_MESSAGES = 1 << PERMISSION_SHIFT_SEND_TTS_MESSAGES
PERMISSION_MASK_MANAGE_MESSAGES = 1 << PERMISSION_SHIFT_MANAGE_MESSAGES
PERMISSION_MASK_EMBED_LINKS = 1 << PERMISSION_SHIFT_EMBED_LINKS
PERMISSION_MASK_ATTACH_FILES = 1 << PERMISSION_SHIFT_ATTACH_FILES
PERMISSION_MASK_READ_MESSAGE_HISTORY = 1 << PERMISSION_SHIFT_READ_MESSAGE_HISTORY
PERMISSION_MASK_MENTION_EVERYONE = 1 << PERMISSION_SHIFT_MENTION_EVERYONE
PERMISSION_MASK_USE_EXTERNAL_EMOJIS = 1 << PERMISSION_SHIFT_USE_EXTERNAL_EMOJIS
PERMISSION_MASK_VIEW_GUILD_INSIGHTS = 1 << PERMISSION_SHIFT_VIEW_GUILD_INSIGHTS
PERMISSION_MASK_CONNECT = 1 << PERMISSION_SHIFT_CONNECT
PERMISSION_MASK_SPEAK = 1 << PERMISSION_SHIFT_SPEAK
PERMISSION_MASK_MUTE_USERS = 1 << PERMISSION_SHIFT_MUTE_USERS
PERMISSION_MASK_DEAFEN_USERS = 1 << PERMISSION_SHIFT_DEAFEN_USERS
PERMISSION_MASK_MOVE_USERS = 1 << PERMISSION_SHIFT_MOVE_USERS
PERMISSION_MASK_USE_VOICE_ACTIVATION = 1 << PERMISSION_SHIFT_USE_VOICE_ACTIVATION
PERMISSION_MASK_CHANGE_NICKNAME = 1 << PERMISSION_SHIFT_CHANGE_NICKNAME
PERMISSION_MASK_MANAGE_NICKNAMES = 1 << PERMISSION_SHIFT_MANAGE_NICKNAMES
PERMISSION_MASK_MANAGE_ROLES = 1 << PERMISSION_SHIFT_MANAGE_ROLES
PERMISSION_MASK_MANAGE_WEBHOOKS = 1 << PERMISSION_SHIFT_MANAGE_WEBHOOKS
PERMISSION_MASK_MANAGE_GUILD_EXPRESSIONS = 1 << PERMISSION_SHIFT_MANAGE_GUILD_EXPRESSIONS
PERMISSION_MASK_USE_APPLICATION_COMMANDS = 1 << PERMISSION_SHIFT_USE_APPLICATION_COMMANDS
PERMISSION_MASK_REQUEST_TO_SPEAK = 1 << PERMISSION_SHIFT_REQUEST_TO_SPEAK
PERMISSION_MASK_MANAGE_EVENTS = 1 << PERMISSION_SHIFT_MANAGE_EVENTS
PERMISSION_MASK_MANAGE_THREADS = 1 << PERMISSION_SHIFT_MANAGE_THREADS
PERMISSION_MASK_CREATE_PUBLIC_THREADS = 1 << PERMISSION_SHIFT_CREATE_PUBLIC_THREADS
PERMISSION_MASK_CREATE_PRIVATE_THREADS = 1 << PERMISSION_SHIFT_CREATE_PRIVATE_THREADS
PERMISSION_MASK_USE_EXTERNAL_STICKERS = 1 << PERMISSION_SHIFT_USE_EXTERNAL_STICKERS
PERMISSION_MASK_SEND_MESSAGES_IN_THREADS = 1 << PERMISSION_SHIFT_SEND_MESSAGES_IN_THREADS
PERMISSION_MASK_USE_EMBEDDED_ACTIVITIES = 1 << PERMISSION_SHIFT_USE_EMBEDDED_ACTIVITIES
PERMISSION_MASK_MODERATE_USERS = 1 << PERMISSION_SHIFT_MODERATE_USERS
PERMISSION_MASK_VIEW_CREATOR_MONETIZATION_ANALYTICS = 1 << PERMISSION_SHIFT_VIEW_CREATOR_MONETIZATION_ANALYTICS
PERMISSION_MASK_USE_SOUNDBOARD = 1 << PERMISSION_SHIFT_USE_SOUNDBOARD
PERMISSION_MASK_CREATE_GUILD_EXPRESSIONS = 1 << PERMISSION_SHIFT_CREATE_GUILD_EXPRESSIONS
PERMISSION_MASK_CREATE_GUILD_EVENTS = 1 << PERMISSION_SHIFT_CREATE_GUILD_EVENTS
PERMISSION_MASK_USE_EXTERNAL_SOUNDS = 1 << PERMISSION_SHIFT_USE_EXTERNAL_SOUNDS
PERMISSION_MASK_SEND_VOICE_MESSAGES = 1 << PERMISSION_SHIFT_SEND_VOICE_MESSAGES
PERMISSION_MASK_USE_CLYDE_AI = 1 << PERMISSION_SHIFT_USE_CLYDE_AI
PERMISSION_MASK_SET_VOICE_CHANNEL_STATUS = 1 << PERMISSION_SHIFT_SET_VOICE_CHANNEL_STATUS
PERMISSION_MASK_SEND_POLLS = 1 << PERMISSION_SHIFT_SEND_POLLS
PERMISSION_MASK_USE_EXTERNAL_APPLICATION_COMMANDS = 1 << PERMISSION_SHIFT_USE_EXTERNAL_APPLICATION_COMMANDS


class Permission(FlagBase):
    """
    Represents a Discord permission.
    
    The implemented permissions are the following:
    
    +---------------------------------------+-------------------+
    | Respective name                       | Bitwise position  |
    +=======================================+===================+
    | create_instant_invite                 |  0                |
    +---------------------------------------+-------------------+
    | kick_users                            |  1                |
    +---------------------------------------+-------------------+
    | ban_users                             |  2                |
    +---------------------------------------+-------------------+
    | administrator                         |  3                |
    +---------------------------------------+-------------------+
    | manage_channels                       |  4                |
    +---------------------------------------+-------------------+
    | manage_guild                          |  5                |
    +---------------------------------------+-------------------+
    | add_reactions                         |  6                |
    +---------------------------------------+-------------------+
    | view_audit_logs                       |  7                |
    +---------------------------------------+-------------------+
    | priority_speaker                      |  8                |
    +---------------------------------------+-------------------+
    | stream                                |  9                |
    +---------------------------------------+-------------------+
    | view_channel                          | 10                |
    +---------------------------------------+-------------------+
    | send_messages                         | 11                |
    +---------------------------------------+-------------------+
    | send_tts_messages                     | 12                |
    +---------------------------------------+-------------------+
    | manage_messages                       | 13                |
    +---------------------------------------+-------------------+
    | embed_links                           | 14                |
    +---------------------------------------+-------------------+
    | attach_files                          | 15                |
    +---------------------------------------+-------------------+
    | read_message_history                  | 16                |
    +---------------------------------------+-------------------+
    | mention_everyone                      | 17                |
    +---------------------------------------+-------------------+
    | use_external_emojis                   | 18                |
    +---------------------------------------+-------------------+
    | connect                               | 20                |
    +---------------------------------------+-------------------+
    | speak                                 | 21                |
    +---------------------------------------+-------------------+
    | mute_users                            | 22                |
    +---------------------------------------+-------------------+
    | deafen_users                          | 23                |
    +---------------------------------------+-------------------+
    | move_users                            | 24                |
    +---------------------------------------+-------------------+
    | use_voice_activation                  | 25                |
    +---------------------------------------+-------------------+
    | change_nickname                       | 26                |
    +---------------------------------------+-------------------+
    | manage_nicknames                      | 27                |
    +---------------------------------------+-------------------+
    | manage_roles                          | 28                |
    +---------------------------------------+-------------------+
    | manage_webhooks                       | 29                |
    +---------------------------------------+-------------------+
    | manage_guild_expressions              | 30                |
    +---------------------------------------+-------------------+
    | use_application_commands              | 31                |
    +---------------------------------------+-------------------+
    | request_to_speak                      | 32                |
    +---------------------------------------+-------------------+
    | manage_events                         | 33                |
    +---------------------------------------+-------------------+
    | manage_threads                        | 34                |
    +---------------------------------------+-------------------+
    | create_public_threads                 | 35                |
    +---------------------------------------+-------------------+
    | create_private_threads                | 36                |
    +---------------------------------------+-------------------+
    | use_external_stickers                 | 37                |
    +---------------------------------------+-------------------+
    | send_messages_in_threads              | 38                |
    +---------------------------------------+-------------------+
    | use_embedded_activities               | 39                |
    +---------------------------------------+-------------------+
    | moderate_users                        | 40                |
    +---------------------------------------+-------------------+
    | view_creator_monetization_analytics   | 41                |
    +---------------------------------------+-------------------+
    | use_soundboard                        | 42                |
    +---------------------------------------+-------------------+
    | create_guild_expressions              | 43                |
    +---------------------------------------+-------------------+
    | create_events                         | 44                |
    +---------------------------------------+-------------------+
    | use_external_sounds                   | 45                |
    +---------------------------------------+-------------------+
    | send_voice_messages                   | 46                |
    +---------------------------------------+-------------------+
    | user_clyde_ai                         | 47                |
    +---------------------------------------+-------------------+
    | set_voice_channel_status              | 48                |
    +---------------------------------------+-------------------+
    | send_polls                            | 49                |
    +---------------------------------------+-------------------+
    | use_external_application_commands     | 50                |
    +---------------------------------------+-------------------+
    
    Each permission can be accessed as property with `can_` + it's respective name, meanwhile a new edited permission
    can be created with the `allow_...` and with the `deny_...` methods.
    """
    create_instant_invite = F(PERMISSION_SHIFT_CREATE_INSTANT_INVITE)
    kick_users = F(PERMISSION_SHIFT_KICK_USERS)
    ban_users = F(PERMISSION_SHIFT_BAN_USERS)
    administrator = F(PERMISSION_SHIFT_ADMINISTRATOR)
    manage_channels = F(PERMISSION_SHIFT_MANAGE_CHANNELS)
    manage_guild = F(PERMISSION_SHIFT_MANAGE_GUILD)
    add_reactions = F(PERMISSION_SHIFT_ADD_REACTION)
    view_audit_logs = F(PERMISSION_SHIFT_VIEW_AUDIT_LOGS)
    priority_speaker = F(PERMISSION_SHIFT_PRIORITY_SPEAKER)
    stream = F(PERMISSION_SHIFT_STREAM)
    view_channel = F(PERMISSION_SHIFT_VIEW_CHANNEL)
    send_messages = F(PERMISSION_SHIFT_SEND_MESSAGES)
    send_tts_messages = F(PERMISSION_SHIFT_SEND_TTS_MESSAGES)
    manage_messages = F(PERMISSION_SHIFT_MANAGE_MESSAGES)
    embed_links = F(PERMISSION_SHIFT_EMBED_LINKS)
    attach_files = F(PERMISSION_SHIFT_ATTACH_FILES)
    read_message_history = F(PERMISSION_SHIFT_READ_MESSAGE_HISTORY)
    mention_everyone = F(PERMISSION_SHIFT_MENTION_EVERYONE)
    use_external_emojis = F(PERMISSION_SHIFT_USE_EXTERNAL_EMOJIS)
    view_guild_insights = F(PERMISSION_SHIFT_VIEW_GUILD_INSIGHTS)
    connect = F(PERMISSION_SHIFT_CONNECT)
    speak = F(PERMISSION_SHIFT_SPEAK)
    mute_users = F(PERMISSION_SHIFT_MUTE_USERS)
    deafen_users = F(PERMISSION_SHIFT_DEAFEN_USERS)
    move_users = F(PERMISSION_SHIFT_MOVE_USERS)
    use_voice_activation = F(PERMISSION_SHIFT_USE_VOICE_ACTIVATION)
    change_nickname = F(PERMISSION_SHIFT_CHANGE_NICKNAME)
    manage_nicknames = F(PERMISSION_SHIFT_MANAGE_NICKNAMES)
    manage_roles = F(PERMISSION_SHIFT_MANAGE_ROLES)
    manage_webhooks = F(PERMISSION_SHIFT_MANAGE_WEBHOOKS)
    manage_guild_expressions = F(PERMISSION_SHIFT_MANAGE_GUILD_EXPRESSIONS)
    use_application_commands = F(PERMISSION_SHIFT_USE_APPLICATION_COMMANDS)
    request_to_speak = F(PERMISSION_SHIFT_REQUEST_TO_SPEAK)
    manage_events = F(PERMISSION_SHIFT_MANAGE_EVENTS)
    manage_threads = F(PERMISSION_SHIFT_MANAGE_THREADS)
    create_public_threads = F(PERMISSION_SHIFT_CREATE_PUBLIC_THREADS)
    create_private_threads = F(PERMISSION_SHIFT_CREATE_PRIVATE_THREADS)
    use_external_stickers = F(PERMISSION_SHIFT_USE_EXTERNAL_STICKERS)
    send_messages_in_threads = F(PERMISSION_SHIFT_SEND_MESSAGES_IN_THREADS)
    use_embedded_activities = F(PERMISSION_SHIFT_USE_EMBEDDED_ACTIVITIES)
    moderate_users = F(PERMISSION_SHIFT_MODERATE_USERS)
    view_creator_monetization_analytics = F(PERMISSION_SHIFT_VIEW_CREATOR_MONETIZATION_ANALYTICS)
    use_soundboard = F(PERMISSION_SHIFT_USE_SOUNDBOARD)
    create_guild_expressions = F(PERMISSION_SHIFT_CREATE_GUILD_EXPRESSIONS)
    create_events = F(PERMISSION_SHIFT_CREATE_GUILD_EVENTS)
    use_external_sounds = F(PERMISSION_SHIFT_USE_EXTERNAL_SOUNDS)
    send_voice_messages = F(PERMISSION_SHIFT_SEND_VOICE_MESSAGES)
    use_clyde_ai = F(PERMISSION_SHIFT_USE_CLYDE_AI)
    set_voice_channel_status = F(PERMISSION_SHIFT_SET_VOICE_CHANNEL_STATUS)
    send_polls = F(PERMISSION_SHIFT_SEND_POLLS)
    use_external_application_commands = F(PERMISSION_SHIFT_USE_EXTERNAL_APPLICATION_COMMANDS)
    
    
    @property
    def can_create_instant_invite(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_create_instant_invite` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.create_instant_invite

    
    @property
    def can_kick_users(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_kick_users` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.kick_users

    
    @property
    def can_ban_users(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_ban_users` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.ban_users

    
    @property
    def can_administrator(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_administrator` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.administrator

    
    @property
    def can_manage_channels(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_channels` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_channels

    
    @property
    def can_manage_guild(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_guild` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_guild

    
    @property
    def can_add_reactions(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_add_reactions` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.add_reactions

    
    @property
    def can_view_audit_logs(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_view_audit_logs` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.view_audit_logs

    
    @property
    def can_priority_speaker(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_priority_speaker` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.priority_speaker

    
    @property
    def can_stream(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_stream` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.stream

    
    @property
    def can_view_channel(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_view_channel` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.view_channel

    
    @property
    def can_send_messages(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_send_messages` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.send_messages

    
    @property
    def can_send_tts_messages(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_send_tts_messages` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.send_tts_messages

    
    @property
    def can_manage_messages(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_messages` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_messages

    
    @property
    def can_embed_links(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_embed_links` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.embed_links

    
    @property
    def can_attach_files(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_attach_files` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.attach_files

    
    @property
    def can_read_message_history(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_read_message_history` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.read_message_history

    
    @property
    def can_mention_everyone(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_mention_everyone` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.mention_everyone

    
    @property
    def can_use_external_emojis(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_external_emojis` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_external_emojis

    
    @property
    def can_view_guild_insights(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_view_guild_insights` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.view_guild_insights

    
    @property
    def can_connect(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_connect` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.connect

    
    @property
    def can_speak(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_speak` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.speak

    
    @property
    def can_mute_users(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_mute_users` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.mute_users

    
    @property
    def can_deafen_users(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_deafen_users` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.deafen_users

    
    @property
    def can_move_users(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_move_users` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.move_users

    
    @property
    def can_use_voice_activation(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_voice_activation` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_voice_activation

    
    @property
    def can_change_nickname(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_change_nickname` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.change_nickname

    
    @property
    def can_manage_nicknames(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_nicknames` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_nicknames

    
    @property
    def can_manage_roles(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_roles` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_roles

    
    @property
    def can_manage_webhooks(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_webhooks` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_webhooks

    
    @property
    def can_manage_guild_expressions(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_guild_expressions` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_guild_expressions

    
    @property
    def can_use_application_commands(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_application_commands` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_application_commands

    
    @property
    def can_request_to_speak(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_request_to_speak` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.request_to_speak

    
    @property
    def can_manage_events(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_events` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_events

    
    @property
    def can_manage_threads(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_manage_threads` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.manage_threads

    
    @property
    def can_create_public_threads(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_create_public_threads` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.create_public_threads

    
    @property
    def can_create_private_threads(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_create_private_threads` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.create_private_threads

    
    @property
    def can_use_external_stickers(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_external_stickers` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_external_stickers

    
    @property
    def can_send_messages_in_threads(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_send_messages_in_threads` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.send_messages_in_threads

    
    @property
    def can_use_embedded_activities(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_embedded_activities` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_embedded_activities

    
    @property
    def can_moderate_users(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_moderate_users` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.moderate_users

    
    @property
    def can_view_creator_monetization_analytics(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_view_creator_monetization_analytics` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.view_creator_monetization_analytics

    
    @property
    def can_use_soundboard(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_soundboard` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_soundboard

    
    @property
    def can_create_guild_expressions(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_create_guild_expressions` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.create_guild_expressions

    
    @property
    def can_create_events(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_create_events` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.create_events

    
    @property
    def can_use_external_sounds(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_external_sounds` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_external_sounds

    
    @property
    def can_send_voice_messages(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_send_voice_messages` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.send_voice_messages

    
    @property
    def can_use_clyde_ai(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_clyde_ai` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_clyde_ai

    
    @property
    def can_set_voice_channel_status(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_set_voice_channel_status` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.set_voice_channel_status

    
    @property
    def can_send_polls(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_send_polls` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.send_polls

    
    @property
    def can_use_external_application_commands(self):
        if DO_CAN_DEPRECATION:
            warn(
                f'`{type(self).__name__}.can_use_external_application_commands` is deprecated and will be removed in 2025 October.',
                FutureWarning,
                stacklevel = 2
            )
        return self.use_external_application_commands

    
    def allow_create_instant_invite(self):
        warn(
            f'`{type(self).__name__}.allow_create_instant_invite` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.create_instant_invite.mask)
    
    
    def allow_kick_users(self):
        warn(
            f'`{type(self).__name__}.allow_kick_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.kick_users.mask)
    
    
    def allow_ban_users(self):
        warn(
            f'`{type(self).__name__}.allow_ban_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.ban_users.mask)
    
    
    def allow_administrator(self):
        warn(
            f'`{type(self).__name__}.allow_administrator` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.administrator.mask)
    
    
    def allow_manage_channels(self):
        warn(
            f'`{type(self).__name__}.allow_manage_channels` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_channels.mask)
    
    
    def allow_manage_guild(self):
        warn(
            f'`{type(self).__name__}.allow_manage_guild` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_guild.mask)
    
    
    def allow_add_reactions(self):
        warn(
            f'`{type(self).__name__}.allow_add_reactions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.add_reactions.mask)
    
    
    def allow_view_audit_logs(self):
        warn(
            f'`{type(self).__name__}.allow_view_audit_logs` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.view_audit_logs.mask)
    
    
    def allow_priority_speaker(self):
        warn(
            f'`{type(self).__name__}.allow_priority_speaker` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.priority_speaker.mask)
    
    
    def allow_stream(self):
        warn(
            f'`{type(self).__name__}.allow_stream` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.stream.mask)
    
    
    def allow_view_channel(self):
        warn(
            f'`{type(self).__name__}.allow_view_channel` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.view_channel.mask)
    
    
    def allow_send_messages(self):
        warn(
            f'`{type(self).__name__}.allow_send_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.send_messages.mask)
    
    
    def allow_send_tts_messages(self):
        warn(
            f'`{type(self).__name__}.allow_send_tts_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.send_tts_messages.mask)
    
    
    def allow_manage_messages(self):
        warn(
            f'`{type(self).__name__}.allow_manage_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_messages.mask)
    
    
    def allow_embed_links(self):
        warn(
            f'`{type(self).__name__}.allow_embed_links` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.embed_links.mask)
    
    
    def allow_attach_files(self):
        warn(
            f'`{type(self).__name__}.allow_attach_files` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.attach_files.mask)
    
    
    def allow_read_message_history(self):
        warn(
            f'`{type(self).__name__}.allow_read_message_history` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.read_message_history.mask)
    
    
    def allow_mention_everyone(self):
        warn(
            f'`{type(self).__name__}.allow_mention_everyone` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.mention_everyone.mask)
    
    
    def allow_use_external_emojis(self):
        warn(
            f'`{type(self).__name__}.allow_use_external_emojis` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_external_emojis.mask)
    
    
    def allow_view_guild_insights(self):
        warn(
            f'`{type(self).__name__}.allow_view_guild_insights` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.view_guild_insights.mask)
    
    
    def allow_connect(self):
        warn(
            f'`{type(self).__name__}.allow_connect` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.connect.mask)
    
    
    def allow_speak(self):
        warn(
            f'`{type(self).__name__}.allow_speak` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.speak.mask)
    
    
    def allow_mute_users(self):
        warn(
            f'`{type(self).__name__}.allow_mute_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.mute_users.mask)
    
    
    def allow_deafen_users(self):
        warn(
            f'`{type(self).__name__}.allow_deafen_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.deafen_users.mask)
    
    
    def allow_move_users(self):
        warn(
            f'`{type(self).__name__}.allow_move_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.move_users.mask)
    
    
    def allow_use_voice_activation(self):
        warn(
            f'`{type(self).__name__}.allow_use_voice_activation` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_voice_activation.mask)
    
    
    def allow_change_nickname(self):
        warn(
            f'`{type(self).__name__}.allow_change_nickname` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.change_nickname.mask)
    
    
    def allow_manage_nicknames(self):
        warn(
            f'`{type(self).__name__}.allow_manage_nicknames` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_nicknames.mask)
    
    
    def allow_manage_roles(self):
        warn(
            f'`{type(self).__name__}.allow_manage_roles` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_roles.mask)
    
    
    def allow_manage_webhooks(self):
        warn(
            f'`{type(self).__name__}.allow_manage_webhooks` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_webhooks.mask)
    
    
    def allow_manage_guild_expressions(self):
        warn(
            f'`{type(self).__name__}.allow_manage_guild_expressions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_guild_expressions.mask)
    
    
    def allow_use_application_commands(self):
        warn(
            f'`{type(self).__name__}.allow_use_application_commands` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_application_commands.mask)
    
    
    def allow_request_to_speak(self):
        warn(
            f'`{type(self).__name__}.allow_request_to_speak` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.request_to_speak.mask)
    
    
    def allow_manage_events(self):
        warn(
            f'`{type(self).__name__}.allow_manage_events` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_events.mask)
    
    
    def allow_manage_threads(self):
        warn(
            f'`{type(self).__name__}.allow_manage_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.manage_threads.mask)
    
    
    def allow_create_public_threads(self):
        warn(
            f'`{type(self).__name__}.allow_create_public_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.create_public_threads.mask)
    
    
    def allow_create_private_threads(self):
        warn(
            f'`{type(self).__name__}.allow_create_private_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.create_private_threads.mask)
    
    
    def allow_use_external_stickers(self):
        warn(
            f'`{type(self).__name__}.allow_use_external_stickers` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_external_stickers.mask)
    
    
    def allow_send_messages_in_threads(self):
        warn(
            f'`{type(self).__name__}.allow_send_messages_in_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.send_messages_in_threads.mask)
    
    
    def allow_use_embedded_activities(self):
        warn(
            f'`{type(self).__name__}.allow_use_embedded_activities` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_embedded_activities.mask)
    
    
    def allow_moderate_users(self):
        warn(
            f'`{type(self).__name__}.allow_moderate_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.moderate_users.mask)
    
    
    def allow_view_creator_monetization_analytics(self):
        warn(
            f'`{type(self).__name__}.allow_view_creator_monetization_analytics` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.view_creator_monetization_analytics.mask)
    
    
    def allow_use_soundboard(self):
        warn(
            f'`{type(self).__name__}.allow_use_soundboard` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_soundboard.mask)
    
    
    def allow_create_guild_expressions(self):
        warn(
            f'`{type(self).__name__}.allow_create_guild_expressions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.create_guild_expressions.mask)
    
    
    def allow_create_events(self):
        warn(
            f'`{type(self).__name__}.allow_create_events` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.create_events.mask)
    
    
    def allow_use_external_sounds(self):
        warn(
            f'`{type(self).__name__}.allow_use_external_sounds` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_external_sounds.mask)
    
    
    def allow_send_voice_messages(self):
        warn(
            f'`{type(self).__name__}.allow_send_voice_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.send_voice_messages.mask)
    
    
    def allow_use_clyde_ai(self):
        warn(
            f'`{type(self).__name__}.allow_use_clyde_ai` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_clyde_ai.mask)
    
    
    def allow_set_voice_channel_status(self):
        warn(
            f'`{type(self).__name__}.allow_set_voice_channel_status` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.set_voice_channel_status.mask)
    
    
    def allow_send_polls(self):
        warn(
            f'`{type(self).__name__}.allow_send_polls` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.send_polls.mask)
    
    
    def allow_use_external_application_commands(self):
        warn(
            f'`{type(self).__name__}.allow_use_external_application_commands` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self | self.use_external_application_commands.mask)
    
    
    def deny_create_instant_invite(self):
        warn(
            f'`{type(self).__name__}.deny_create_instant_invite` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.create_instant_invite.mask)
    
    
    def deny_kick_users(self):
        warn(
            f'`{type(self).__name__}.deny_kick_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.kick_users.mask)
    
    
    def deny_ban_users(self):
        warn(
            f'`{type(self).__name__}.deny_ban_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.ban_users.mask)
    
    
    def deny_administrator(self):
        warn(
            f'`{type(self).__name__}.deny_administrator` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.administrator.mask)
    
    
    def deny_manage_channels(self):
        warn(
            f'`{type(self).__name__}.deny_manage_channels` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_channels.mask)
    
    
    def deny_manage_guild(self):
        warn(
            f'`{type(self).__name__}.deny_manage_guild` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_guild.mask)
    
    
    def deny_add_reactions(self):
        warn(
            f'`{type(self).__name__}.deny_add_reactions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.add_reactions.mask)
    
    
    def deny_view_audit_logs(self):
        warn(
            f'`{type(self).__name__}.deny_view_audit_logs` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.view_audit_logs.mask)
    
    
    def deny_priority_speaker(self):
        warn(
            f'`{type(self).__name__}.deny_priority_speaker` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.priority_speaker.mask)
    
    
    def deny_stream(self):
        warn(
            f'`{type(self).__name__}.deny_stream` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.stream.mask)
    
    
    def deny_view_channel(self):
        warn(
            f'`{type(self).__name__}.deny_view_channel` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.view_channel.mask)
    
    
    def deny_send_messages(self):
        warn(
            f'`{type(self).__name__}.deny_send_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.send_messages.mask)
    
    
    def deny_send_tts_messages(self):
        warn(
            f'`{type(self).__name__}.deny_send_tts_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.send_tts_messages.mask)
    
    
    def deny_manage_messages(self):
        warn(
            f'`{type(self).__name__}.deny_manage_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_messages.mask)
    
    
    def deny_embed_links(self):
        warn(
            f'`{type(self).__name__}.deny_embed_links` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.embed_links.mask)
    
    
    def deny_attach_files(self):
        warn(
            f'`{type(self).__name__}.deny_attach_files` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.attach_files.mask)
    
    
    def deny_read_message_history(self):
        warn(
            f'`{type(self).__name__}.deny_read_message_history` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.read_message_history.mask)
    
    
    def deny_mention_everyone(self):
        warn(
            f'`{type(self).__name__}.deny_mention_everyone` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.mention_everyone.mask)
    
    
    def deny_use_external_emojis(self):
        warn(
            f'`{type(self).__name__}.deny_use_external_emojis` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_external_emojis.mask)
    
    
    def deny_view_guild_insights(self):
        warn(
            f'`{type(self).__name__}.deny_view_guild_insights` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.view_guild_insights.mask)
    
    
    def deny_connect(self):
        warn(
            f'`{type(self).__name__}.deny_connect` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.connect.mask)
    
    
    def deny_speak(self):
        warn(
            f'`{type(self).__name__}.deny_speak` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.speak.mask)
    
    
    def deny_mute_users(self):
        warn(
            f'`{type(self).__name__}.deny_mute_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.mute_users.mask)
    
    
    def deny_deafen_users(self):
        warn(
            f'`{type(self).__name__}.deny_deafen_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.deafen_users.mask)
    
    
    def deny_move_users(self):
        warn(
            f'`{type(self).__name__}.deny_move_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.move_users.mask)
    
    
    def deny_use_voice_activation(self):
        warn(
            f'`{type(self).__name__}.deny_use_voice_activation` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_voice_activation.mask)
    
    
    def deny_change_nickname(self):
        warn(
            f'`{type(self).__name__}.deny_change_nickname` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.change_nickname.mask)
    
    
    def deny_manage_nicknames(self):
        warn(
            f'`{type(self).__name__}.deny_manage_nicknames` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_nicknames.mask)
    
    
    def deny_manage_roles(self):
        warn(
            f'`{type(self).__name__}.deny_manage_roles` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_roles.mask)
    
    
    def deny_manage_webhooks(self):
        warn(
            f'`{type(self).__name__}.deny_manage_webhooks` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_webhooks.mask)
    
    
    def deny_manage_guild_expressions(self):
        warn(
            f'`{type(self).__name__}.deny_manage_guild_expressions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_guild_expressions.mask)
    
    
    def deny_use_application_commands(self):
        warn(
            f'`{type(self).__name__}.deny_use_application_commands` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_application_commands.mask)
    
    
    def deny_request_to_speak(self):
        warn(
            f'`{type(self).__name__}.deny_request_to_speak` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.request_to_speak.mask)
    
    
    def deny_manage_events(self):
        warn(
            f'`{type(self).__name__}.deny_manage_events` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_events.mask)
    
    
    def deny_manage_threads(self):
        warn(
            f'`{type(self).__name__}.deny_manage_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.manage_threads.mask)
    
    
    def deny_create_public_threads(self):
        warn(
            f'`{type(self).__name__}.deny_create_public_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.create_public_threads.mask)
    
    
    def deny_create_private_threads(self):
        warn(
            f'`{type(self).__name__}.deny_create_private_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.create_private_threads.mask)
    
    
    def deny_use_external_stickers(self):
        warn(
            f'`{type(self).__name__}.deny_use_external_stickers` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_external_stickers.mask)
    
    
    def deny_send_messages_in_threads(self):
        warn(
            f'`{type(self).__name__}.deny_send_messages_in_threads` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.send_messages_in_threads.mask)
    
    
    def deny_use_embedded_activities(self):
        warn(
            f'`{type(self).__name__}.deny_use_embedded_activities` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_embedded_activities.mask)
    
    
    def deny_moderate_users(self):
        warn(
            f'`{type(self).__name__}.deny_moderate_users` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.moderate_users.mask)
    
    
    def deny_view_creator_monetization_analytics(self):
        warn(
            f'`{type(self).__name__}.deny_view_creator_monetization_analytics` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.view_creator_monetization_analytics.mask)
    
    
    def deny_use_soundboard(self):
        warn(
            f'`{type(self).__name__}.deny_use_soundboard` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_soundboard.mask)
    
    
    def deny_create_guild_expressions(self):
        warn(
            f'`{type(self).__name__}.deny_create_guild_expressions` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.create_guild_expressions.mask)
    
    
    def deny_create_events(self):
        warn(
            f'`{type(self).__name__}.deny_create_events` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.create_events.mask)
    
    
    def deny_use_external_sounds(self):
        warn(
            f'`{type(self).__name__}.deny_use_external_sounds` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_external_sounds.mask)
    
    
    def deny_send_voice_messages(self):
        warn(
            f'`{type(self).__name__}.deny_send_voice_messages` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.send_voice_messages.mask)
    
    
    def deny_use_clyde_ai(self):
        warn(
            f'`{type(self).__name__}.deny_use_clyde_ai` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_clyde_ai.mask)
    
    
    def deny_set_voice_channel_status(self):
        warn(
            f'`{type(self).__name__}.deny_set_voice_channel_status` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.set_voice_channel_status.mask)
    
    
    def deny_send_polls(self):
        warn(
            f'`{type(self).__name__}.deny_send_polls` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.send_polls.mask)
    
    
    def deny_use_external_application_commands(self):
        warn(
            f'`{type(self).__name__}.deny_use_external_application_commands` is deprecated and will be removed in 2025 April.',
            FutureWarning,
            stacklevel = 2
        )
        return int.__new__(type(self), self & ~self.use_external_application_commands.mask)
    
    


    

PERMISSION_ALL = Permission().update_by_keys(
    create_instant_invite = True,
    kick_users = True,
    ban_users = True,
    administrator = True,
    manage_channels = True,
    manage_guild = True,
    add_reactions = True,
    view_audit_logs = True,
    priority_speaker = True,
    stream = True,
    view_channel = True,
    send_messages = True,
    send_tts_messages = True,
    manage_messages = True,
    embed_links = True,
    attach_files = True,
    read_message_history = True,
    mention_everyone = True,
    use_external_emojis = True,
    view_guild_insights = True,
    connect = True,
    speak = True,
    mute_users = True,
    deafen_users = True,
    move_users = True,
    use_voice_activation = True,
    change_nickname = True,
    manage_nicknames = True,
    manage_roles = True,
    manage_webhooks = True,
    manage_guild_expressions = True,
    use_application_commands = True,
    request_to_speak = True,
    manage_events = True,
    manage_threads = True,
    create_public_threads = True,
    create_private_threads = True,
    use_external_stickers = True,
    send_messages_in_threads = True,
    use_embedded_activities = True,
    moderate_users = True,
    view_creator_monetization_analytics = True,
    use_soundboard = True,
    create_guild_expressions = True,
    create_events = True,
    use_external_sounds = True,
    send_voice_messages = True,
    use_clyde_ai = True,
    set_voice_channel_status = True,
    send_polls = True,
    use_external_application_commands = True,
)

PERMISSION_NONE = Permission()

PERMISSION_PRIVATE = Permission().update_by_keys(
    create_instant_invite = False,
    kick_users = False,
    ban_users = False,
    administrator = False,
    manage_channels = False,
    manage_guild = False,
    add_reactions = True,
    view_audit_logs = False,
    priority_speaker = False,
    stream = False,
    view_channel = True,
    send_messages = True,
    send_tts_messages = False,
    manage_messages = False,
    embed_links = True,
    attach_files = True,
    read_message_history = True,
    mention_everyone = True,
    use_external_emojis = True,
    view_guild_insights = False,
    connect = False,
    speak = False,
    mute_users = False,
    deafen_users = False,
    move_users = False,
    use_voice_activation = True,
    change_nickname = False,
    manage_nicknames = False,
    manage_roles = False,
    manage_webhooks = False,
    manage_guild_expressions = False,
    manage_events = False,
    manage_threads = False,
    use_application_commands = True,
    request_to_speak = False,
    use_external_stickers = True,
    send_messages_in_threads = False,
    use_embedded_activities = False,
    moderate_users = False,
    view_creator_monetization_analytics = False,
    use_soundboard = True,
    create_guild_expressions = False,
    create_events = False,
    use_external_sounds = True,
    send_voice_messages = True,
    use_clyde_ai = False,
    set_voice_channel_status = False,
    send_polls = True,
    use_external_application_commands = True,
)

PERMISSION_PRIVATE_BOT = PERMISSION_PRIVATE.update_by_keys(
    read_message_history = False,
)

PERMISSION_GROUP = PERMISSION_PRIVATE.update_by_keys(
    create_instant_invite = True,
)

PERMISSION_GROUP_OWNER = PERMISSION_GROUP.update_by_keys(
    kick_users = True,
)

PERMISSION_TEXT_ALL = Permission().update_by_keys(
    create_instant_invite = False,
    kick_users = False,
    ban_users = False,
    administrator = False,
    manage_channels = False,
    manage_guild = False,
    add_reactions = True,
    view_audit_logs = False,
    priority_speaker = False,
    stream = False,
    view_channel = False,
    send_messages = True,
    send_tts_messages = True,
    manage_messages = False,
    embed_links = True,
    attach_files = True,
    read_message_history = False,
    mention_everyone = True,
    use_external_emojis = True,
    view_guild_insights = False,
    connect = False,
    speak = False,
    mute_users = False,
    deafen_users = False,
    move_users = False,
    use_voice_activation = False,
    change_nickname = False,
    manage_nicknames = False,
    manage_roles = False,
    manage_webhooks = False,
    manage_guild_expressions = False,
    use_application_commands = True,
    request_to_speak = False,
    manage_events = False,
    manage_threads = True,
    create_public_threads = True,
    create_private_threads = True,
    use_external_stickers = True,
    send_messages_in_threads = True,
    use_embedded_activities = False,
    moderate_users = False,
    view_creator_monetization_analytics = False,
    use_soundboard = False,
    create_guild_expressions = False,
    create_events = False,
    use_external_sounds = False,
    send_voice_messages = True,
    use_clyde_ai = True,
    set_voice_channel_status = False,
    send_polls = True,
    use_external_application_commands = True,
)

PERMISSION_TEXT_DENY = Permission(~PERMISSION_TEXT_ALL)

PERMISSION_DENY_SEND_MESSAGES_IN_THREADS_ONLY = PERMISSION_ALL.update_by_keys(
    send_messages_in_threads = False,
)

PERMISSION_DENY_SEND_MESSAGES_ONLY = PERMISSION_ALL.update_by_keys(
    send_messages = False,
)

PERMISSION_DENIED_FOR_GUILD_VOICE = Permission(~0).update_by_keys(
    request_to_speak = False,
)

PERMISSION_DENIED_FOR_GUILD_STAGE = Permission(~0).update_by_keys(
    set_voice_channel_status = False,
)

PERMISSION_VOICE_ONLY = Permission().update_by_keys(
    create_instant_invite = False,
    kick_users = False,
    ban_users = False,
    administrator = False,
    manage_channels = False,
    manage_guild = False,
    add_reactions = False,
    view_audit_logs = False,
    priority_speaker = True,
    stream = True,
    view_channel = False,
    send_messages = False,
    send_tts_messages = False,
    manage_messages = False,
    embed_links = False,
    attach_files = False,
    read_message_history = False,
    mention_everyone = False,
    use_external_emojis = False,
    view_guild_insights = False,
    connect = True,
    speak = True,
    mute_users = True,
    deafen_users = True,
    move_users = True,
    use_voice_activation = True,
    change_nickname = False,
    manage_nicknames = False,
    manage_roles = False,
    manage_webhooks = False,
    manage_guild_expressions = False,
    use_application_commands = False,
    request_to_speak = True,
    manage_events = True,
    manage_threads = False,
    create_public_threads = False,
    create_private_threads = False,
    use_external_stickers = False,
    send_messages_in_threads = False,
    use_embedded_activities = True,
    moderate_users = False,
    view_creator_monetization_analytics = False,
    use_soundboard = True,
    create_guild_expressions = False,
    create_events = False,
    use_external_sounds = True,
    send_voice_messages = True,
    use_clyde_ai = False,
    set_voice_channel_status = False,
    send_polls = False,
    use_external_application_commands = False,
)

PERMISSION_VOICE_DENY = Permission(~PERMISSION_VOICE_ONLY)

PERMISSION_THREAD_AND_VOICE_DENY = PERMISSION_VOICE_DENY.update_by_keys(
    create_public_threads = False,
    create_private_threads = False,
    use_external_stickers = False,
)

PERMISSION_VOICE_DENY_CONNECTION = PERMISSION_VOICE_DENY.update_by_keys(
    manage_roles = False,
    manage_channels = False,
)

PERMISSION_TEXT_AND_VOICE_DENY = Permission(PERMISSION_THREAD_AND_VOICE_DENY | PERMISSION_VOICE_DENY)

PERMISSION_STAGE_MODERATOR = Permission().update_by_keys(
    create_instant_invite = False,
    kick_users = False,
    ban_users = False,
    administrator = False,
    manage_channels = False,
    manage_guild = False,
    add_reactions = False,
    view_audit_logs = False,
    priority_speaker = False,
    stream = False,
    view_channel = False,
    send_messages = False,
    send_tts_messages = False,
    manage_messages = False,
    embed_links = False,
    attach_files = False,
    read_message_history = False,
    mention_everyone = False,
    use_external_emojis = False,
    view_guild_insights = False,
    connect = False,
    speak = False,
    mute_users = True,
    deafen_users = False,
    move_users = True,
    use_voice_activation = False,
    change_nickname = False,
    manage_nicknames = False,
    manage_roles = False,
    manage_webhooks = False,
    manage_guild_expressions = False,
    use_application_commands = False,
    request_to_speak = True,
    manage_events = True,
    manage_threads = False,
    create_public_threads = False,
    create_private_threads = False,
    use_external_stickers = False,
    send_messages_in_threads = False,
    moderate_users = False,
    view_creator_monetization_analytics = False,
    use_soundboard = False,
    create_guild_expressions = False,
    create_events = False,
    use_external_sounds = False,
    use_clyde_ai = False,
    set_voice_channel_status = False,
    send_polls = False,
    use_external_application_commands = False,
)

PERMISSION_CAN_SEND_MESSAGES_ALL = Permission().update_by_keys(
    send_messages = True,
    send_messages_in_threads = True,
)
